/*********************************************************************
 * FILE: %Z%%M%  %I%  %W% %G% %U%
 *
 * Description:
 *
 * Servlet that handles "forgot password".
 * Note that in order for a user to be able to use this feature, he
 * must have previously defined his challenge-response answers.
 *********************************************************************/

package examples.expi;

import java.io.*;
import java.util.*;
import java.rmi.RemoteException;

import javax.servlet.*;
import javax.servlet.http.*;
import javax.security.auth.Subject;
import javax.security.auth.login.*;

import com.ibm.itim.apps.*;
import com.ibm.itim.apps.identity.*;
import com.ibm.itim.apps.jaas.spi.PlatformPrincipal;
import com.ibm.itim.apps.jaas.callback.PlatformCallbackHandler;
import com.ibm.itim.apps.provisioning.*;
import com.ibm.itim.dataservices.model.DistinguishedName;
import com.ibm.itim.dataservices.model.domain.Account;
import com.ibm.itim.apps.provisioning.AccountMO;

public class ForgotPasswordServlet extends HttpServlet {

	public static String LOGON = "logon.jsp";
	public static final String FORGOT_PWD = "forgotpwd.jsp";
	public static final String FORGOT_PWD_INFO = "forgotpwdinfo.jsp";
	public static final String USERID = "userid";
	public static final String ERROR_MESSAGE = "errorMessage";
	public static final String INFO_MESSAGE = "infoMessage";
	public static final String CHALLENGES = "challenges";
	public static final String PLATFORM_CALLBACK_HANDLER =
		"platformCallbackHandler";
	public static final String PLATFORM_CONTEXT = "platform";

	// The following get set in init()
	private static String userIDAttribute;
	private static boolean changeOnlyTIMPassword;
	private static boolean displayPassword;
	private static boolean ssoEnabled;

	/**
	 * Read in the properties file.
	 * @see javax.servlet.GenericServlet#init()
	 */

	public void init() throws ServletException {
		expiUtil eu = new expiUtil();

		// Are we to only change the TIM Account password, or change all?
		String par = eu.getProperty(expiUtil.CHANGEONLYITIMPASSWORD).trim();
		this.changeOnlyTIMPassword = par.compareToIgnoreCase("true") == 0;
		log("init(): changeOnlyTIMPassword = " + this.changeOnlyTIMPassword);

		// Are we to display or email generated passwords?
		par = eu.getProperty(expiUtil.DISPLAYPASSWORD).trim();
		this.displayPassword = par.compareToIgnoreCase("true") == 0;
		log("init(): displayPassword = " + this.displayPassword);

		// Is SSO enabled?
		LOGON = eu.getPropertySSOCheck(eu.LOGON_PAGE);
		ssoEnabled = eu.isSSOEnabled();
		/*
		 * If SSO is enabled, use the "username" attribute.
		 * Otherwise use whatever's set in the expiUtil.
		 */
		if (ssoEnabled)
			this.userIDAttribute = "username";
		else
			this.userIDAttribute = expiUtil.LOGON_ID;
		log("init(): userIDAttribute = " + this.userIDAttribute);
		log("ForgotPasswordServlet.init() complete");

	} // init

	/**
	 * Method destroy
	 * Right now this does nothing.
	 * @see javax.servlet.Servlet#destroy()
	 */

	public void destroy() {

	} // destroy

	/**
	 * Method doGet
	 * This just calls doPost().
	 * @see javax.servlet.http.HttpServlet#doGet(HttpServletRequest, HttpServletResponse)
	 */
	public void doGet(HttpServletRequest request, HttpServletResponse response)
		throws ServletException, IOException {

		log("ForgotPasswordServlet.doGet() forwarding to doPost()");
		doPost(request, response);

	} // doGet

	/**
	 * @see javax.servlet.http.HttpServlet#doPost(HttpServletRequest, HttpServletResponse)
	 */
	public void doPost(
		HttpServletRequest request,
		HttpServletResponse response)
		throws ServletException, IOException {

		log("ForgotPasswordServlet.doPost()");

		// Need a session  -- at minimum to do error handling
		HttpSession session = request.getSession(true);

		// Find out how we got here and figure out what to do!
		// Note: no action specified means to get the challenges.
		String action = request.getParameter("action");
		log("ForgotPasswordServlet.doPost(): action = " + action);

		if (isNullOrEmpty(action)) {
			if (!session.isNew()) {
				log("[ForgotPasswordServlet.doPost:] invalidating previous session");
				session.invalidate();
				session = request.getSession();
			}
			doGetChallenges(request, response, session);
		} else if (action.equals("authenticate")) {
			doAuthenticate(request, response, session);
		} else {
			// Unknown action
		}

		return;

	} // doPost

	/**
	 * Method doAuthenticate.
	 * Attempt to authenticate the user via challenge response.
	 * @param request
	 * @param response
	 * @param session
	 * @throws ServletException
	 * @throws IOException
	 */
	private void doAuthenticate(
		HttpServletRequest request,
		HttpServletResponse response,
		HttpSession session)
		throws ServletException, IOException {

		// First check responses -- this will forward back to jsp
		Map cars = validateResponses(request, response, session);
		if ((cars == null) || cars.isEmpty())
			return;

		// Get the PlatformCallbackHandler and set the responses on it
		PlatformCallbackHandler pch =
			(PlatformCallbackHandler) session.getAttribute(
				PLATFORM_CALLBACK_HANDLER);
		if (pch == null) {
			log("doAuthenticate(): No PlatformCallbackHandler in the session");
			goToPage(LOGON, null, session, request, response);
			return;
		}
		pch.setResponses(cars);

		// Attempt to login; if it succeeds change the password!
		try {
			LoginContext lc = new LoginContext("ITIM", pch);
			lc.login();
			log("doAuthenticate(): LoginContext.login() succeeded");
			changePassword(request, response, session, lc.getSubject());
		} catch (LoginException e) {
			log(
				"doAuthenticate(): LoginContext.login() failed with error"
					+ e.getMessage());

			String eMsg = e.getMessage();
			if (isNullOrEmpty(eMsg)) {
				goToPage(
					FORGOT_PWD,
					"An unknown error occurred.",
					session,
					request,
					response);
			} else if (
				eMsg.equals(
					"com.ibm.itim.apps.ejb.home.HomeBean.INVALID_ANSWERS")) {
				goToPage(
					FORGOT_PWD,
					"One or more of the responses was incorrect.",
					session,
					request,
					response);
			} else {
				goToPage(
					FORGOT_PWD,
					e.getMessage(),
					session,
					request,
					response);
			}
		}

		return;
	}

	/**
	 * Method changePassword.
	 * @param request
	 * @param response
	 * @param session
	 * @param subject
	 * @throws ServletException
	 * @throws IOException
	 */
	private void changePassword(
		HttpServletRequest request,
		HttpServletResponse response,
		HttpSession session,
		Subject subject)
		throws ServletException, IOException {

		// First off, get a PasswordManager
		PlatformContext platform =
			(PlatformContext) session.getAttribute(PLATFORM_CONTEXT);
		if (platform == null) {
			log("changePassword(): No PlatformContext!");
			goToPage(
				LOGON,
				"Unable to reset password at this time.  Please try again later.",
				session,
				request,
				response);
			return;

		}

		PasswordManager pm = new PasswordManager(platform, subject);
		if (pm == null) {
			log("changePassword(): PasswordManager is empty!!!!");
			goToPage(
				LOGON,
				"Unable to complete the challenge response process at this this time.  Please try again later.",
				session,
				request,
				response);
			return;
		}

		// Next lookup the PersonMO;  need this to get the accounts
		expiUtil expiu = new expiUtil();
		PersonMO pmo = expiu.lookupPerson(platform, subject);
		if (pmo == null) {
			log("ChangePasswordServlet.changePassword(): PersonMO is null");
			goToPage(
				LOGON,
				"Unable to complete the challenge response process at this this time.  Please try again later.",
				session,
				request,
				response);
			return;
		}

		/*
		 * Get the accounts; this will check to see if we're changing only
		 * the TIM account; also sets up any forwarding
		 */
		Collection accounts =
			getAccountsToChange(session, request, response, pm, pmo);
		log("changePassword(): Num accounts to change = " + accounts.size());
		if ((accounts == null) || accounts.isEmpty())
			return;

		try {
			String password = pm.generatePassword(accounts);
			BulkRequest br =
				pm.changePassword(accounts, password, new Date(), true);

			String message =
				"A password change request has been submitted to the server.<br>Request ID is: "
					+ br.getID()
					+ "<br><br>";

			if (displayPassword) {
				message =
					message.concat(
						"You will be notified via email when the password change is complete.  <br>Your new password will be:  "
							+ password);
			} else {
				message =
					message.concat(
						"You will be notified via email when the password change is complete.  <br>The email message will contain the new password.");
			}

			goToInfoPage(FORGOT_PWD_INFO, message, session, request, response);

		} catch (RemoteException e) {
			// Unable to communicate with the platform
			log("changePassword(): Caught RemoteException");
			goToPage(
				FORGOT_PWD,
				"Unable to communicate with the server.  Please try again later.",
				session,
				request,
				response);

		} catch (AuthorizationException e) {
			// User is unauthorized to change the password of any of his accounts
			log("changePassword(): Caught AuthorizationException");
			goToPage(
				LOGON,
				"You cannot use the \"Forgot Password\" feature because you are not authrized to change the password of any of your accounts.  Please contact an administrator.",
				session,
				request,
				response);

		} catch (PasswordRuleException e) {
			// Password rules could not be merged
			log("changePassword(): Caught PasswordRuleException");
			goToPage(
				LOGON,
				"You cannot use the \"Forgot Password\" feature because a new password could not be generated due to merge conflicts with the password rules.  Please contact an administrator.",
				session,
				request,
				response);

		} catch (InvalidPasswordException e) {
			/*
				 * The generated password is invalid?  Not likely unless the rules
				 * were changed mid-stream!
			 */
			log("changePassword(): Caught InvalidPasswordException");
			goToPage(
				LOGON,
				"Unable to complete challenge response because a new password cannot be generated.  Please try again later.",
				session,
				request,
				response);

		} catch (ApplicationException e) {
			// Password could not be generated; account may have been removed
			log("changePassword(): Caught ApplicationException");
			goToPage(
				LOGON,
				"Unable to complete challenge response because a new password cannot be generated.  Please try again later.",
				session,
				request,
				response);

		}

		return;

	} // changePassword

	/**
	 * Method validateResponses.
	 * Check to make sure that there was an answer for each response.
	 * Setup the forward if there were any problems.
	 * @param request
	 * @param response
	 * @param session
	 * @return Map
	 * @throws ServletException
	 * @throws IOException
	 */

	private Map validateResponses(
		HttpServletRequest request,
		HttpServletResponse response,
		HttpSession session)
		throws ServletException, IOException {

		Map cars = (Map) session.getAttribute(CHALLENGES);
		Iterator it = cars.keySet().iterator();

		while (it.hasNext()) {
			String challenge = (String) it.next();
			log("e(): challenge: " + challenge);
			String responseToChallenge =
				(String) request.getParameter(challenge);
			log("validateResponses(): response: " + responseToChallenge);
			if (isNullOrEmpty(responseToChallenge)) {
				goToPage(
					FORGOT_PWD,
					"One or more responses was empty.  You must provide a response for each challenge.",
					session,
					request,
					response);
				return null;
			} else {
				cars.put(challenge, responseToChallenge);
			}
		}

		return cars;
	}

	/**
	 * Method doGetChallenges.
	 * Get the challenges for the userid that was specified in
	 * the request, set them in the session, and forward to
	 * forgotpwd.jsp.
	 * @param request
	 * @param response
	 * @param session
	 * @throws ServletException
	 * @throws IOException
	 */

	private void doGetChallenges(
		HttpServletRequest request,
		HttpServletResponse response,
		HttpSession session)
		throws ServletException, IOException {

		log("doGetChallenges() begin");

		// Get the userID from the request; if there wasn't one
		// this will handle setting up the forward back to the logon page.
		String userID = getUserID(request, response, session);
		if (userID == null)
			return;

		Map challenges = null;
		try {
			// First off, setup PlatformCallbackHandler
			PlatformCallbackHandler pch =
				new PlatformCallbackHandler(userID, Locale.US);
			expiUtil eu = new expiUtil();
			PlatformContext platform = eu.createPlatformContext();
			pch.setPlatformContext(platform);
			challenges = pch.getChallenges();

			session.setAttribute(CHALLENGES, challenges);
			session.setAttribute(PLATFORM_CALLBACK_HANDLER, pch);
			session.setAttribute(PLATFORM_CONTEXT, platform);
			session.setAttribute(expiUtil.LOGON_ID, userID);

			/*
			 * See if there are challenges.
			 *
			 * NOTE: If challenges come back empty but there are no errors, we
			 * can't tell if it's because C/R is disabled or if the user
			 * just doesn't have any challenges defined.
			 * The API that we could use to find out requires a Subject.
			 */
			if ((challenges == null) || challenges.isEmpty()) {
				goToPage(
					LOGON,
					"Either you do not have any challenges configured, or challenge-resonse is disabled.",
					session,
					request,
					response);
			} else {
				goToPage(FORGOT_PWD, null, session, request, response);
			}
		} catch (ApplicationException e) {
			log(
				"getChallenges(): Caught ApplicationException: "
					+ e.getMessage());
			goToPage(LOGON, e.getMessage(), session, request, response);
		} catch (RemoteException e) {
			log("getChallenges(): Caught RemoteException: " + e.getMessage());
			goToPage(LOGON, e.getMessage(), session, request, response);
		}

		log("doGetChallenges() complete");

	} // getChallenges

	/**
	 * Method getUserID.
	 * Get the userID from the request based on the setting of
	 * userIDAttribute.  If there isn't one, setup the forward back
	 * to the logon page.
	 *
	 * Note: the logon page checks to make sure that the userID has a
	 * value.  But the user might have tried to get here directly.
	 * @param request
	 * @param response
	 * @param session
	 * @return String
	 * @throws ServletException
	 * @throws IOException
	 */

	private String getUserID(
		HttpServletRequest request,
		HttpServletResponse response,
		HttpSession session)
		throws ServletException, IOException {

		String userID = (String) request.getParameter(this.userIDAttribute);

		if (userID == null) {
			goToPage(
				LOGON,
				"Enter your User ID and then click on the \"Forgot Your Password\" link",
				session,
				request,
				response);
		}

		log("ForgotPasswordServlet.getUserID(): userID = " + userID);

		return userID;

	} // getUserID

	/**
	 * Method goToPage.
	 * Set the message in the request and forward it on to the specified page.
	 * @param page
	 * @param errorMessage
	 * @param session
	 * @param request
	 * @param response
	 * @throws ServletException
	 * @throws IOException
	 */

	private void goToPage(
		String page,
		String errorMessage,
		HttpSession session,
		HttpServletRequest request,
		HttpServletResponse response)
		throws ServletException, IOException {

		session.setAttribute(ERROR_MESSAGE, errorMessage);

		RequestDispatcher dispatcher =
			getServletContext().getRequestDispatcher("/" + page);

		dispatcher.forward(request, response);

	} // goToPage

	/**
	 * Method goToInfoPage.
	 * Set the message in the request and forward it on to the specified page.
	 * @param page
	 * @param infoMessage
	 * @param session
	 * @param request
	 * @param response
	 * @throws ServletException
	 * @throws IOException
	 */

	private void goToInfoPage(
		String page,
		String infoMessage,
		HttpSession session,
		HttpServletRequest request,
		HttpServletResponse response)
		throws ServletException, IOException {

		session.setAttribute(INFO_MESSAGE, infoMessage);
		if (ssoEnabled) {
			request.setAttribute("ERROR", infoMessage);
		}

		RequestDispatcher dispatcher =
			getServletContext().getRequestDispatcher("/" + page);

		dispatcher.forward(request, response);

	} // goToPage

	/**
	 * Method isNullOrEmpty.
	 * Returns true if the given object is null or equals "".
	 * @param o
	 * @return boolean
	 */
	private boolean isNullOrEmpty(Object o) {

		if ((o == null) || (o.equals("")))
			return true;
		else
			return false;

	} // isNullOrEmpty

	/**
	 * Method validateSessionData.
	 * Validates the session data.  If anything is amiss, forward back
	 * to the logon page.
	 * @param session
	 * @param request
	 * @param response
	 * @param subject
	 * @param platform
	 * @return boolean
	 * @throws ServletException
	 * @throws IOException
	 */

	private boolean validateSessionData(
		HttpSession session,
		HttpServletRequest request,
		HttpServletResponse response,
		Subject subject,
		PlatformContext platform)
		throws ServletException, IOException {

		log("ForgotPasswordServlet.validateSessionData(): start");

		boolean isValid = true;

		if ((subject == null) || (platform == null)) {
			goToPage(
				LOGON,
				"Your session is invalid, please login again.",
				session,
				request,
				response);
			isValid = false;
		}

		return isValid;

	} // validateSessionData

	/**
	 * Method getAccountsToChange.
	 *
	 * Consults the changeOnlyTIMPassword property and retrieves the
	 * appropriate accounts.
	 * @param session
	 * @param request
	 * @param response
	 * @param pm
	 * @param pmo
	 * @return Collection
	 * @throws ServletException
	 * @throws IOException
	 */
	private Collection getAccountsToChange(
		HttpSession session,
		HttpServletRequest request,
		HttpServletResponse response,
		PasswordManager pm,
		PersonMO pmo)
		throws ServletException, IOException {

		log("getAccountsToChange(): start");

		Collection accounts = null;

		try {
			accounts = pm.getPasswordAccounts(pmo);

			if (changeOnlyTIMPassword == true) {
				accounts = getTIMAccount(session, request, response, accounts);
			}

			// Setup the forward if needed
			if ((accounts == null) || accounts.isEmpty()) {
				log("getAccountsToChange(): Accounts is null or no Accounts found");
				goToPage(
					LOGON,
					"You have answered the challenge questions correctly, however, your password cannot be changed because your accounts cannot be retrieved. Please try again later.",
					session,
					request,
					response);
			}

		} catch (ApplicationException e) {

			/*
			 * Thrown if unable to obtain the accounts.
			 * This may possibly be caused by the subject being removed by another
			 * client previous to this call.  Note that since the user is logged
			 * in, they should have at least one account!
			 */
			log(
				"getAccountsToChange(): Caught ApplicationException: "
					+ e.getMessage());
			goToPage(
				LOGON,
				"You have answered the challenge questions correctly, however, your password cannot be changed because your accounts cannot be retrieved.  Please try again later.",
				session,
				request,
				response);

		} catch (RemoteException e) {

			log(
				"ChangePasswordServlet.getAccountsToChange(): Caught RemoteException: "
					+ e.getMessage());

			// Thrown if unable to communicate with platform.
			goToPage(
				FORGOT_PWD,
				"Unable to retrieve your accounts.  The server may be busy or there may be a more serious problem.  Please try again.  If the problem persists, contact your system administrator.",
				session,
				request,
				response);
		}

		return accounts;

	} // getAccountsToChange

	/**
	 * Method getTIMAccount.
	 * Returns a Collection that contains the TIM Account of the user
	 * who is trying to reset his password.  Note that the user could
	 * have more than one TIM Account, that's why we have to do the
	 * check.
	* @param session
	 * @param request
	 * @param response
	 * @param accounts
	 * @return Collection
	 * @throws ServletException
	 * @throws IOException
	 */
	private Collection getTIMAccount(
		HttpSession session,
		HttpServletRequest request,
		HttpServletResponse response,
		Collection accounts)
		throws ServletException, IOException {

		String userID = (String) session.getAttribute(expiUtil.LOGON_ID);
		log("getTIMAccount(): user ID = " + userID);
		if (userID == null) {
			log("getTIMAccount(): could not get userID from session");
			return null;
		}

		Collection onlyTimAccount = new Vector();
		Iterator it = accounts.iterator();

		try {
			while (it.hasNext()) {
				AccountMO amo = (AccountMO) it.next();

				/*
				 * Note: SERVICE_ENROLE_PROFILE_NAME = "ITIMService"
				 * This is the service name of the account.
				 * If you were searching for an Solaris account, you would
				 * put that service name here instead.
				 */
				log(
					"getTIMAccount(): Profile name "
						+ amo.getService().getData().getServiceProfileName());
				if (amo
					.getService()
					.getData()
					.getServiceProfileName()
					.equals(expiUtil.SERVICE_ENROLE_PROFILE_NAME)) {
					log(
						"getTIMAccount(): AccountMO user ID: "
							+ amo.getData().getUserId());
					String aUserID = amo.getData().getUserId();
					if (aUserID.equalsIgnoreCase(userID)) {
						onlyTimAccount.add(amo);
						return onlyTimAccount;
					}
				}
			} // while
		} catch (Exception e) {
			accounts = null;
		}

		if (onlyTimAccount == null) {
			log("getTIMAccount(): Could not find TIM Account for user.");
		}

		return onlyTimAccount;

	} // getTIMAccount

} // ForgotPasswordServlet
